Program
    Uses
        cobra2D,
        keyset,
        poxml,
        svg

Const
    SUPPORTING_STARS = 16


Var
    ms_start, ms, fm, cfps: Integer
    fd: Real = 1.0
    
Procedure SyncFramer()
Begin
    fm = Millisecs - ms
    ms = Millisecs
    fd = (fm / 1000.0) * 60.0
    fd = SmoothFrameDelta(fd) 
End

Function CalcFps(m1: Integer, m2: Integer) : Integer
Begin

    If ((m2 - m1) > 0) then
        result = (1000.0/(m2 - m1))
    Else
        result = 999
    Endif

End

Function SmoothFrameDelta(fd: Real) : Real ; export
Var
    avg, fd1, fd2, fd3, fd4, fd5, fd6, fd7, fd8, fd9: Static Real
Begin

    avg = (fd1 + fd2 + fd3 + fd4 + fd5 + fd6 + fd7 + fd8 + fd9 + fd) / 10.0 
    
    fd1 = fd2
    fd2 = fd3
    fd3 = fd4
    fd4 = fd5
    fd5 = fd6
    fd6 = fd7
    fd7 = fd8
    fd8 = fd9
    fd9 = fd
    
    If fd > avg then fd = avg
    result = fd

End

Var
    i, j, k, m: Integer = 0
    mySvg: ^svgs
    
    canvas: Element
    
    zoom: Real = 1.0
    star: ^svgPolygons
    x, y, s: Real
    xdir, ydir: Integer = 1
    
    supportingStar: Array [SUPPORTING_STARS] of ^svgs
    
Begin

    SetAppName("SVG Demo")
    OpenScreen(640, 480, 32, FALSE, COB_SHOWBORDER)
    VSync(1)
    Randomize(Millisecs)

    For i = 0 to (SUPPORTING_STARS)
        supportingStar[i] = LoadSvg('star_simple.svg')
        SetSvgScale(supportingStar[i], 0.2, 0.2)
        SetSvgPosition(supportingStar[i], 30 + (600/SUPPORTING_STARS) * i, Rand(30, 450))
    Next
    
    mySvg = LoadSvg('star_detailed.svg')
    mySvg.angle = 0
    SetSvgPosition(mySvg, 320, 240)
    star = GetPolygonById(mySvg, 'star')
    
    canvas = CreateSprite(640, 480)
    PenColor(ToRGBA(0,0,0), canvas)
    SetFont("verdana", 8, 0, canvas)
    
    SyncFramer()
    While (KeyDown(VK_ESCAPE) = FALSE)
        SyncFramer()
    
        CLS(ToRGBA(255,255,255), canvas)
        
        m = Millisecs
        
        For i = 0 to (SUPPORTING_STARS)
            SetSvgAngle(supportingStar[i], (m/10) mod 360)
            supportingStar[i].xscale = 0.1 + (0.2 * (Abs(Sin(m/15))))
            supportingStar[i].yscale = supportingStar[i].xscale 
        Next
        
        mySvg.angle = (m/40) mod 360
        mySvg.xscale = 0.5 + Abs(Sin(m/100))
        mySvg.yscale = mySvg.xscale
        
        s = Sin(m/5) * 30
        For i = 0 to (star.totalPoints-1)
            x = PolyX(star.oPoints, i)
            y = PolyY(star.oPoints, i)
            SetPolyX(star.points, i, x + (s * Sin(y)))
            star.updated = TRUE
        Next
        
        If star.miny < 0 then ydir = Abs(ydir)
        If star.maxy > 480 then ydir = -Abs(ydir)
        If star.minx < 0 then xdir = Abs(xdir)
        If star.maxx > 640 then xdir = -Abs(xdir)
        
        If fd > 0 then
            SetSvgPosition(mySvg, mySvg.x + (2.0 * fd * xdir), mySvg.y + (2.0 * fd * ydir))
        Endif  
        
        RenderSvgs(canvas) 
        
        Text(10, 10, "VSync enabled. FPS: "+cfps, canvas)
        Text(10, 25, "Real-time SVG Tech Demo", canvas)
        Text(10, 40, "2010-09-04", canvas)
        Text(10, 55, "tophatstuff.co.uk", canvas)
    
        Flip
        //Pause(1)
        cfps = CalcFps(ms, Millisecs)
    Wend

End