{
    
    Poxml - A "Plain Old Xml" parser for the Cobra programming language.
    
    ---------------------------------------
    
    Updates: http://www.tophatstuff.co.uk/?p=103
    
    ---------------------------------------
    
    Poxml should be able to parse any valid XML that exists*. However,
    it's only a simple "plain old xml" parser; doctypes, directives and
    namespaces are all ignored. 
    
    Poxml is, however, very strict in what it will accept. If you try
    to parse an improperly formatted file, Poxml won't try (and fail) to
    second-guess you. Instead, it will stop parsing and tell you exactly
    what to fix.
         
    
    * (except that Poxml currently doesn't support single quotes as
       delimiters, only double. This should be fixed in a later release.)
    
    ---------------------------------------   
    
    Comparason between Poxml and the native Cobra XML library:
        
    Unlike Cobra's native XML...
        
        Poxml tells you where your errors are!
        Poxml properly supports <!-- --> comments
        Poxml supports any arbitrary line breaks and whitespace between
              delimiters.
        Poxml preserves linespaces in node values
        Poxml actually works reliably!
        
    Unlike Poxml...
        
        Cobra XML can use both double quotes and single quotes as
              delimiters (poxml can't at this time)
        Cobra will continue to parse invalid XML. Depending on your
              point of view, this is either a bug or a feature!
        Cobra contains functionality for saving XML.
        
    ---------------------------------------
    
    Exported functions:
    
        PoxLoad(filename): ^poxFiles
        PoxRootNode(poxFile): ^poxNodes
        PoxChild(poxNode, index): ^poxNodes
        PoxChildIndexByName(poxNode, name): Integer (-1 for not found)
        PoxChildren(poxNode): Integer
        PoxNodeValue(poxNode): String
        PoxParameter(poxNode, index): String
        PoxParameterIndexByName(poxNode, name): Integer (-1 for not found)
        PoxParameters: Integer
        PoxFree(poxFile)  
    
    ---------------------------------------
    
    Copyright (c) 2010 Ben Golightly

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.
    
    --------------------------------------- 

}
Unit

Const

    // Version
    POX_VERSION = 20100903 ; export

    // Config
    SUPPRESS_ERRORS         = FALSE // supresses error messageboxes
    MAX_CHILD_NODES         = 1024 // Max nesting

    // Tokens
    TOKEN_INVALID           = 0
    TOKEN_STRING            = 1 // does not include quotes
    TOKEN_EQUALS            = 2
    TOKEN_QUOTE             = 3 // " (todo: single quotes)
    TOKEN_IDENT             = 4 // variable names
    TOKEN_LITERAL           = 5 // e.g. <foo>literal</foo>
    TOKEN_OPEN_TAG          = 6 // <
    TOKEN_CLOSE_TAG         = 7 // >    
    TOKEN_OPEN_CLOSING_TAG  = 8 // </
    TOKEN_EMPTY_TERMINATIOR = 9 // />, terminatior for an empty element, e.g. in the case of <br /> 
    TOKEN_OPEN_COMMENT      = 10
    TOKEN_CLOSE_COMMENT     = 11
    TOKEN_OPEN_DIRECTIVE    = 12
    TOKEN_CLOSE_DIRECTIVE   = 13
    TOKEN_WHITESPACE        = 14
    TOKEN_SLASH             = 15
    TOKEN_END_STREAM        = 16 // End of token stream
    
    // Lexer States
    LEXER_STATE_NO_STR      = 0
    LEXER_STATE_IN_STR      = 1
    LEXER_STATE_END_STR     = 2
    LEXER_STATE_NO_TAG      = 0
    LEXER_STATE_IN_TAG      = 1
    LEXER_STATE_END_TAG     = 2
    LEXER_STATE_NO_COMMENT  = 0
    LEXER_STATE_IN_COMMENT  = 1
    LEXER_STATE_END_COMMENT = 2
    LEXER_STATE_NO_DIRECTIVE  = 0
    LEXER_STATE_IN_DIRECTIVE  = 1
    LEXER_STATE_END_DIRECTIVE = 2

// Types used locally:

Type xStrings = Record
    s: String = ""
EndType ; export

Type poxAttributes = Record
    name : ^xStrings
    value: ^xStrings    
EndType ; export

Type poxNodes = Record
    name            : ^xStrings
    value           : ^xStrings
    attribList      : List of poxAttributes
    attribListCount : Integer = 0 
    nodeList        : List of poxNodes
    nodeListCount   : Integer = 0
EndType ; export

Type poxFiles = Record
    root: ^poxNodes
EndType ; export

// Types used by parser

Type poxSourceLines = Record
    s : String = ""
EndType

Var
    poxSourceList   : List of poxSourceLines
    poxFileList     : List of poxFiles
    poxCurrLine     : ^poxSourceLines
    
    poxLexer_index0             : Integer = 0
    poxLexer_index1             : Integer = 0
    poxLexer_currToken          : Integer
    poxLexer_currLexeme         : String = ""
    poxLexer_currStringState    : Integer
    poxLexer_currCommentState   : Integer
    poxLexer_currTagState       : Integer
    poxLexer_currDirState       : Integer
    
    poxLexer_nodeStack          : Array [MAX_CHILD_NODES] of ^poxNodes
    poxLexer_nodeStackIndex     : Integer = 0
    
    poxErrors: Integer

Function PoxLoad(filename: String) : ^poxFiles ; export
Begin
    result = Pox_Load(filename)
End

Procedure PoxFree(poxFile: ^poxFiles) ; export
Begin
    Pox_Free(poxFile)
End

Procedure Pox_Free(poxFile: ^poxFiles) ; export
Begin
    //If ItemInList(poxFileList, poxFile) = -1 then exit // Doesn't work!
    
    Pox_FreeNode(poxFile.root)  
    Free(poxFile)
End

Procedure Pox_FreeNode(node: ^poxNodes)
Var
    child: ^poxNodes
    attrib: ^poxAttributes
    magic: Boolean
Begin

    While node.nodeListCount > 0
        child = FirstItem(node.nodeList)
        Dec(node.nodeListCount)
        magic = TRUE // Fix for extremely weird bug
        Pox_FreeNode(child)
    Wend
    

    Free(node.name)
    Free(node.value)
    
    While node.attribListCount > 0
        attrib = FirstItem(node.attribList)
        Dec(node.attribListCount) 
        Free(attrib.name)
        Free(attrib.value)
        Free(attrib)
    Wend
    
    Free(node) // This Line throws "Error: List does not exist" in debug
End

Function PoxRootNode(poxFile: ^poxFiles) : ^poxNodes ; export
Begin
    result = poxFile.root
End

Function PoxChildren(poxNode: ^poxNodes) : Integer ; export
Begin
    result = poxNode.nodeListCount
End

Function PoxChild(poxNode: ^poxNodes, i: Integer) : ^poxNodes ; export
Begin
    result = IndexedItem(poxNode.nodeList, i)
End

Function PoxNodeValue(poxNode: ^poxNodes) : String ; export
Begin
    result = poxNode.value.s
End

Function PoxParameter(poxNode: ^poxNodes, i: Integer) : String ; export
Var
    attrib: ^poxAttributes
Begin
    attrib = IndexedItem(poxNode.attribList, i)
    result = attrib.value.s
End

Function PoxParameters(poxNode: ^poxNodes) : Integer ; export
Begin
    result = poxNode.attribListCount
End

Function PoxChildIndexByName(poxNode: ^poxNodes, s: String) : Integer ; export
Var
    child: ^poxNodes
    i: Integer = 0
Begin
    result = -1
    
    Loop child through poxNode.nodeList
        If child.name.s = s then
            result = i
            exit
        Endif
        Inc(i)
    EndLoop
End

Function PoxParameterIndexByName(poxNode: ^poxNodes, s: String) : Integer ; export
Var
    attrib: ^poxAttributes
    i: Integer = 0
Begin
    result = -1
    
    Loop attrib through poxNode.attribList
        If attrib.name.s = s then
            result = i
            exit
        Endif
        Inc(i)
    EndLoop
End

Function Pox_Load(filename: String) : ^poxFiles ; export
Var
    m: Integer
    s: String
    i: Integer
    
    poxSourceLine: ^poxSourceLines
    poxFile: ^poxFiles
    
    activeNode: ^poxNodes
    activeAttribute: ^poxAttributes
    
    bracketOpen: Boolean = FALSE
    bracketClosing: Boolean = FALSE
    emptyTerminator: Boolean = FALSE
    attributeExists: Boolean = FALSE
    nodeIdentCount: Integer = 0
    
    mf: Element
    
    bytes: Integer
    
Begin
    m = Millisecs
    
    If FileExists(filename) = FALSE then
        Pox_Error("File doesn't exist at '"+filename+"'")
        exit
    Endif
    
    mf = CreateMemFile()
    LoadToMemFile(mf, filename)
    If mf = NULL then
        Pox_Error("NULL input!")
        exit
    Endif
    
    // Strip whitespace & comments
    While Not EOF(mf)
        s = ReadLn(mf) + Chr(10)    
        poxSourceLine = NewItem(poxSourceList)
        poxSourceLine.s = s
    Wend
    Closefile(mf)
    s = ""
    
    // Resets
    Pox_ResetLexer()
    poxErrors = 0
    
    // Init
    poxFile = NewItem(poxFileList)
    result = poxFile
    
    // Root node
    New(poxFile.root)
    activeNode = Pox_PushNode(Pox_NewRootNode(poxFile.root))
    
    
    While (TRUE)
        If (Pox_GetNextToken() = TOKEN_END_STREAM) then break
        
        Case poxLexer_currToken of
            
            (TOKEN_WHITESPACE):
            (TOKEN_OPEN_COMMENT):
            (TOKEN_CLOSE_COMMENT):
            (TOKEN_OPEN_DIRECTIVE):
            (TOKEN_CLOSE_DIRECTIVE):
            
            (TOKEN_OPEN_TAG):
                If bracketOpen = TRUE then Pox_ExpectedError(">"); break
                bracketOpen = TRUE
                nodeIdentCount = 0
                attributeExists = FALSE
            
            (TOKEN_EMPTY_TERMINATIOR):
                emptyTerminator = TRUE
            
            (TOKEN_CLOSE_TAG):
                If bracketOpen = FALSE then Pox_UnexpectedError(">"); break
                If nodeIdentCount = 0 then Pox_ExpectedError("identifier") ; break
                If emptyTerminator or bracketClosing then activeNode = Pox_PopNode()
                bracketOpen = FALSE
                bracketClosing = FALSE
                emptyTerminator = FALSE
            
            (TOKEN_OPEN_CLOSING_TAG):
                If bracketOpen = TRUE then Pox_ExpectedError(">"); break
                If (Pox_GetNextToken() <> TOKEN_SLASH) then Pox_ExpectedError("/") ; break
                bracketOpen = TRUE
                bracketClosing = TRUE
                nodeIdentCount = 0
            
            (TOKEN_IDENT):
                If bracketOpen = FALSE then Pox_UnexpectedError("identifier '"+poxLexer_currLexeme+"'") ; break
                Inc(nodeIdentCount)
                If nodeIdentCount = 1 then
                    // <name>
                    If bracketClosing then
                        If (activeNode.name.s <> poxLexer_currLexeme) then
                            Pox_UnexpectedError("</"+poxLexer_currLexeme+">")
                            break
                        Endif
                    Else
                        activeNode = Pox_PushNode(Pox_AddNode(activeNode, poxLexer_currLexeme))
                    Endif
                Else
                    // <name attribute="">
                    If bracketClosing then
                        Pox_UnexpectedError("attribute in closing tag")
                    Else
                        activeAttribute = Pox_AddAttribute(activeNode, poxLexer_currLexeme)
                        attributeExists = TRUE
                    Endif                    
                Endif
            
            (TOKEN_EQUALS):
                If attributeExists = FALSE then Pox_UnexpectedError("="); break
                While Pox_GetNextToken() = TOKEN_WHITESPACE
                Wend
                If poxLexer_currToken <> TOKEN_QUOTE then Pox_ExpectedError("value") ; break
                If Pox_GetNextToken() <> TOKEN_STRING then Pox_ExpectedError("value") ; break
                
                activeAttribute.value.s = Pox_SpecialChrs(poxLexer_currLexeme)
                
                If Pox_GetNextToken() <> TOKEN_QUOTE then Pox_ExpectedError("closing quote") ; break
            
            (TOKEN_LITERAL):
                If bracketOpen = TRUE then Pox_UnexpectedError("literal"); break
                If poxLexer_nodeStackIndex = 1 then Pox_UnexpectedError("literal in global scope"); break
                activeNode.value.s = activeNode.value.s + Pox_SpecialChrs(poxLexer_currLexeme)
                 
            (TOKEN_INVALID):
                Pox_UnexpectedError("unkown '"+poxLexer_currLexeme+"'"); break
            Default:
                Pox_UnexpectedError("(warning) unkown token type "+poxLexer_currToken+" '"+poxLexer_currLexeme+"'")
        EndCase
        
    Wend
    
    If (poxErrors = 0) and (poxLexer_nodeStackIndex > 1) then
        Pox_Error("Error: "+(poxLexer_nodeStackIndex-1)+" node(s) unclosed.")
    Endif    
    
    If (poxErrors > 0) then
        Pox_Error("Aborted reading "+filename+" due to "+poxErrors+" error(s).")
    Endif
    
    Loop poxSourceLine through poxSourceList
        Free(poxSourceLine)
    EndLoop
        
    poxErrors = 0 
End

Function Pox_PushNode(node: ^poxNodes) : ^poxNodes
Begin
    poxLexer_nodeStack[poxLexer_nodeStackIndex] = node
    Inc(poxLexer_nodeStackIndex)
    
    If poxLexer_nodeStackIndex = MAX_CHILD_NODES then
        Pox_Error("FATAL: Stack overflow - poxLexer_nodeStackIndex exceeds maximum of "+MAX_CHILD_NODES+"")
    Endif
    
    result = node
End

Function Pox_PopNode() : ^poxNodes
Begin
    
    If poxLexer_nodeStackIndex = 0 then
        Pox_Error("FATAL: Stack underflow (poxLexer_nodeStackIndex is already zero)")
        exit
    Endif
    
    Dec(poxLexer_nodeStackIndex)    
    result = poxLexer_nodeStack[poxLexer_nodeStackIndex-1] 
End

Function Pox_AddAttribute(node: ^poxNodes, name: String) : ^poxAttributes
Var
    poxAttribute: ^poxAttributes
Begin
    Inc(node.attribListCount)
    poxAttribute = NewItem(node.attribList)
    New(poxAttribute.name)
    New(poxAttribute.value)
    poxAttribute.name.s = name
    result = poxAttribute
End

Function Pox_AddNode(parent: ^poxNodes, name: String): ^poxNodes
Var
    child: ^poxNodes
Begin
    
    // Create new
    Inc(parent.nodeListCount)
    child = NewItem(parent.nodeList)
    New(child.name)
    New(child.value)
    child.name.s = name
    result = child
    
End

Function Pox_NewRootNode(node: ^poxNodes): ^poxNodes
Begin
    
    // Create new
    New(node.name)
    New(node.value)
    node.name.s = ""
    node.value.s = ""
    result = node
    
End

Function Pox_GetNextToken() : Integer
Var
    l, i: Integer
    c: String
    literal: String = ""
    continue: Boolean = FALSE
    iterations: Static Integer = 0
    lastQuote: Static Integer = -123
Begin
    Inc(iterations)

    poxLexer_index0 = poxLexer_index1
    result = TOKEN_INVALID
    
    // Advance to next Line as necessary
    l = Length(poxCurrLine.s)
    If poxLexer_index0 >= l then
        If Pox_LexerNextLine() = FALSE then
            result = TOKEN_END_STREAM
            exit
        Endif
    Endif
    
    If (poxLexer_currStringState = LEXER_STATE_END_STR) then poxLexer_currStringState = LEXER_STATE_NO_STR
    If (poxLexer_currTagState    = LEXER_STATE_END_TAG) then poxLexer_currTagState = LEXER_STATE_NO_TAG
    If (poxLexer_currCommentState = LEXER_STATE_END_COMMENT) then poxLexer_currCommentState = LEXER_STATE_NO_COMMENT
    If (poxLexer_currDirState = LEXER_STATE_END_DIRECTIVE) then poxLexer_currDirState = LEXER_STATE_NO_DIRECTIVE

    // Scan through all potential whitespace    
    If (poxLexer_currStringState <> LEXER_STATE_IN_STR) then
    
        While (TRUE)
            c = Mid(poxCurrLine.s, poxLexer_index0+1, 1)
            If Pox_CharIsWhiteSpace(c) = FALSE then break
            Inc(poxLexer_index0)
        Wend
        
    Endif
    
    poxLexer_index1 = poxLexer_index0
    
    
    
    // Scan through lexeme until a delimiter is met, incrementing index1    
    While (TRUE)
        Conditions
        (poxLexer_currStringState = LEXER_STATE_IN_STR):
            continue = FALSE
        
            l = Length(poxCurrLine.s)
        
            // Strings cannot span multiple lines
            If poxLexer_index1 >= l then
                poxLexer_currToken = TOKEN_INVALID
                result = poxLexer_currToken
                exit 
            Endif
            
            c = Mid(poxCurrLine.s, poxLexer_index1+1, 1)
            
            // Backslash
            If c = '\' then
                poxLexer_index1 = poxLexer_index1 + 2
                continue = TRUE
            Endif  
            
            // quote delimiter
            If (continue = FALSE) and (c = '"') then
                break            
            Endif
            
            If (continue = FALSE) then Inc(poxLexer_index1)
            
            
        (poxLexer_currCommentState = LEXER_STATE_IN_COMMENT):
        
            l = Length(poxCurrLine.s)
        
            // Advance through lines as neccessary
            If poxLexer_index1 >= l-1 then
                If Pox_LexerNextLine() = FALSE then
                    result = TOKEN_END_STREAM
                    exit
                Endif
            Endif
            
            c = Mid(poxCurrLine.s, poxLexer_index1+1, 1)
            
            If c = '>' then
                poxLexer_index0 = poxLexer_index1
                break
            Endif
            
            Inc(poxLexer_index1)


        (poxLexer_currDirState = LEXER_STATE_IN_DIRECTIVE):
        
            l = Length(poxCurrLine.s)
        
            // Advance through lines as neccessary
            If poxLexer_index1 >= l-1 then
                If Pox_LexerNextLine() = FALSE then
                    result = TOKEN_END_STREAM
                    exit
                Endif
            Endif
            
            c = Mid(poxCurrLine.s, poxLexer_index1+1, 2)
            
            If (c = '?>') then // ends <?xml?>
                poxLexer_index0 = poxLexer_index1+1
                poxLexer_index1 = poxLexer_index1+1
                break
            Endif
            
            Inc(poxLexer_index1)            
            
            
        (poxLexer_currTagState = LEXER_STATE_NO_TAG):
        
            l = Length(poxCurrLine.s)
        
            // Advance through lines as neccessary
            If poxLexer_index1 >= l then
                If Pox_LexerNextLine() = FALSE then
                    result = TOKEN_END_STREAM
                    exit
                Endif
            Endif
            
            c = Mid(poxCurrLine.s, poxLexer_index1+1, 1)
            
            If c = '<' then
                break
            Else
                If Length(literal) >= 1 then
                    literal = literal + c
                Else
                    If Pox_CharIsWhiteSpace(c) = FALSE then literal = literal + c
                Endif
            Endif
            
            Inc(poxLexer_index1)
        
        
        Default:
            
            l = Length(poxCurrLine.s)
            If poxLexer_index1 >= l then break
            
            c = Mid(poxCurrLine.s, poxLexer_index1+1, 1)
            
            If Pox_CharIsDelim(c) then break
            
            Inc(poxLexer_index1)
        
        EndConditions       
    Wend
    
    If (poxLexer_index1 - poxLexer_index0) = 0 then Inc(poxLexer_index1)
    poxLexer_currLexeme = ""
    poxLexer_currToken = TOKEN_INVALID
    
    
    
    // Lexeme Isolatted, lies between index0 and index1
    For i = poxLexer_index0 to poxLexer_index1-1
    
        If (poxLexer_currStringState = LEXER_STATE_IN_STR) then
            If(Mid(poxCurrLine.s, i+1, 1) = '\') then Inc(i)
        Endif
        
        poxLexer_currLexeme = poxLexer_currLexeme + Mid(poxCurrLine.s, i+1, 1)
    Next 
    
    l = Length(poxLexer_currLexeme)
    
        
    // Token Identification
    
    // A string token
    If (l > 1) or (Left(poxLexer_currLexeme, 1) <> '"') then // (... <> '"') is for for single Chr strings
        If (poxLexer_currStringState = LEXER_STATE_IN_STR) then
            poxLexer_currToken = TOKEN_STRING
            result = poxLexer_currToken
            exit
        Endif
    Else
        // empty string! There's no lexeme content, so pretend there is then rewind back to the closing quote!
        If (lastQuote = iterations-1) then 
            If (poxLexer_currStringState = LEXER_STATE_IN_STR) then
                poxLexer_currLexeme = ""
                poxLexer_currToken = TOKEN_STRING
                result = poxLexer_currToken
                Dec(poxLexer_index1)
                lastQuote = -123
                exit
            Endif
        Endif        
    Endif    
    
    If Pox_StringIsWhiteSpace(poxLexer_currLexeme) then
        poxLexer_currToken = TOKEN_WHITESPACE
        result = poxLexer_currToken 
        exit
    Endif
    
    
    If l = 1 then
        // Single-character tokens
        
        Conditions
            
            (poxLexer_currLexeme = '/'):
                If (Pox_LexerLookAhead() = ">") then
                    poxLexer_currToken = TOKEN_EMPTY_TERMINATIOR
                Else
                    poxLexer_currToken = TOKEN_SLASH
                Endif
            
            (poxLexer_currLexeme = '"'):
                Case poxLexer_currStringState of
                    (LEXER_STATE_NO_STR):  poxLexer_currStringState = LEXER_STATE_IN_STR
                    (LEXER_STATE_IN_STR):  poxLexer_currStringState = LEXER_STATE_END_STR
                EndCase
                poxLexer_currToken = TOKEN_QUOTE
                lastQuote = iterations 
                
            (poxLexer_currLexeme = '<'):
                Conditions
                    (Pox_LexerLookAhead() = "!"):
                        poxLexer_currToken = TOKEN_OPEN_COMMENT
                        poxLexer_currCommentState = LEXER_STATE_IN_COMMENT
                    (Pox_LexerLookAhead() = "?"):
                        poxLexer_currToken = TOKEN_OPEN_DIRECTIVE
                        poxLexer_currDirState = LEXER_STATE_IN_DIRECTIVE
                    (Pox_LexerLookAhead() = "/"):
                        poxLexer_currToken = TOKEN_OPEN_CLOSING_TAG 
                        poxLexer_currTagState = LEXER_STATE_IN_TAG
                    Default:
                        poxLexer_currToken = TOKEN_OPEN_TAG
                        poxLexer_currTagState = LEXER_STATE_IN_TAG
                EndConditions
            
            (poxLexer_currLexeme = '>'):
                Conditions
                    (poxLexer_currDirState = LEXER_STATE_IN_DIRECTIVE):
                        poxLexer_currToken = TOKEN_CLOSE_DIRECTIVE
                        poxLexer_currDirState = LEXER_STATE_END_COMMENT
                    (poxLexer_currCommentState = LEXER_STATE_IN_COMMENT):
                        poxLexer_currToken = TOKEN_CLOSE_COMMENT
                        poxLexer_currCommentState = LEXER_STATE_END_COMMENT
                    Default:
                        poxLexer_currToken = TOKEN_CLOSE_TAG
                        poxLexer_currTagState = LEXER_STATE_END_TAG
                EndConditions                                       
                            
            (poxLexer_currLexeme = '='):            poxLexer_currToken = TOKEN_EQUALS
            (Pox_CharIsIdent(poxLexer_currLexeme)): poxLexer_CurrToken = TOKEN_IDENT
            
        EndConditions
        
        
    Else
        // Multi-character tokens
        
        Conditions
            (Pox_CharIsIdent(poxLexer_currLexeme)):         poxLexer_CurrToken = TOKEN_IDENT
        EndConditions
        
    Endif
    
    If (poxLexer_currTagState = LEXER_STATE_NO_TAG) and (poxLexer_currCommentState = LEXER_STATE_NO_COMMENT) and (poxLexer_currDirState = LEXER_STATE_NO_DIRECTIVE) then
        If Length(literal) = 0 then
            poxLexer_currToken = TOKEN_WHITESPACE
            result = poxLexer_currToken
            exit 
        Endif
        poxLexer_currToken = TOKEN_LITERAL
        result = poxLexer_currToken
        poxLexer_currLexeme = literal
        literal = ""
        exit 
    Endif
    
    result = poxLexer_currToken
    
End

Function Pox_LexerNextLine() : Boolean
Begin

    If poxCurrLine = LastItem(poxSourceList) then result = FALSE ; exit
    
    poxCurrLine = NextItem(poxSourceList, poxCurrLine)
    
    poxLexer_currStringState = LEXER_STATE_NO_STR
    poxLexer_index0 = 0
    poxLexer_index1 = 0
    
    result = TRUE

End

Function Pox_LexerLookAhead() : String
Var
    index: Integer
    c: String = ""
    tmp_poxCurrLine : ^poxSourceLines
Begin
    result = ""

    index = poxLexer_index1
    tmp_poxCurrLine = poxCurrLine
    
    
    If (poxLexer_currStringState <> LEXER_STATE_IN_STR)
        While (TRUE)
        
            If (index >= Length(poxCurrLine.s)) then
                If poxCurrLine = LastItem(poxSourceList) then
                    poxCurrLine = tmp_poxCurrLine
                    exit
                Else
                    poxCurrLine = NextItem(poxSourceList, poxCurrLine)
                Endif
                
                index = 0
            Endif
            
            c = Mid(poxCurrLine.s, index+1, 1)
            If (Pox_CharIsWhiteSpace(c) = FALSE) then break
            Inc(index)
                    
        Wend
    Endif
    
    result = c    

    // Reset line
    poxCurrLine = tmp_poxCurrLine
    
End

Procedure Pox_ResetLexer()
Begin
    poxCurrLine = FirstItem(poxSourceList)
    poxLexer_index0 = 0
    poxLexer_index1 = 0
    poxLexer_currLexeme = ""
    poxLexer_currToken = TOKEN_INVALID
    poxLexer_currStringState = LEXER_STATE_NO_STR
    poxLexer_currTagState = LEXER_STATE_NO_TAG
    poxLexer_currCommentState = LEXER_STATE_NO_COMMENT
    poxLexer_currDirState = LEXER_STATE_NO_DIRECTIVE
    poxLexer_nodeStackIndex = 0
End

Function Pox_CharIsNumeric(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c)    
    If (a >= 48) and (a <= 57) then result = TRUE else result = FALSE    
End

Function Pox_CharIsHex(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c)
    result = FALSE    
    If (a >= 48) and (a <= 57) then result = TRUE ; exit
    If (a >= 65) and (a <= 70) then result = TRUE ; exit
    If (a >= 97) and (a <= 102) then result = TRUE ; exit    
End

Function Pox_CharIsIdent(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c)
    result = FALSE    
    If (a >= 48) and (a <= 57) then result = TRUE  ; exit // 0 - 9
    If (a >= 65) and (a <= 90) then result = TRUE  ; exit // A - Z
    If (a >= 97) and (a <= 122) then result = TRUE ; exit // a - z
    If (a = 58) // :
    If (a = 46) then result = TRUE // .
    If (a = 95) then result = TRUE // _    
End

Function Pox_CharIsDelim(c: String) : Boolean
Begin
    result = FALSE
    If (c = '"') then result = TRUE  ; exit
    If (c = '=') then result = TRUE  ; exit
    If (c = '<') then result = TRUE  ; exit
    If (c = '/') then result = TRUE  ; exit
    If (c = '>') then result = TRUE  ; exit
    result = Pox_CharIsWhitespace(c)  
End

Function Pox_CharIsWhitespace(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c) // Tab = 9, Space = 32    
    If (a = 9) or (a = 32) or (a = 10) then result = TRUE else result = FALSE
End    

Function Pox_StringIsWhitespace(s: String) : Boolean ; export
Var
    l, i: Integer
    c: String
Begin
    result = FALSE
    l = Length(s)
    If l < 1 then result = TRUE; exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (Pox_CharIsWhitespace(c)) then exit 
    Next
    
    result = TRUE        
End

Function Pox_StringIsInt(s: String) : Boolean ; export
Var
    l, i: Integer
    c: String
Begin
    result = FALSE
    l = Length(s)
    If l < 1 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (Pox_CharIsNumeric(c) or (c = '-')) then exit
        If (i <> 1) and (c = '-') then exit 
    Next
    
    result = TRUE        
End

Function Pox_StringIsFloat(s: String) : Boolean ; export
Var
    l, i: Integer
    radix: Integer = 0
    c: String
Begin
    result = FALSE
    l = Length(s)
    If l < 1 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (Pox_CharIsNumeric(c) or (c = '-') or (c = '.')) then exit
        If (i <> 1) and (c = '-') then exit
        
        If (c = '.') then
            Inc(radix)
            If i = l then exit
        Endif
        If radix > 1 then exit 
    Next
    
    result = TRUE        
End

Function Pox_StringIsIdent(s: String) : Boolean
Var
    l, i: Integer
    c: String
Begin
    result = FALSE
    l = Length(s)
    If l < 1 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If (i = 1) and Pox_CharIsNumeric(c) then exit
        If Pox_CharIsIdent(c) = FALSE then exit 
    Next
    
    result = TRUE        
End

Function Pox_StringIsHex(s: String) : Boolean ; export
Var
    l, i: Integer
    c: String
    sCount: Integer = 0 // symbolCount
Begin
    result = FALSE
    l = Length(s)
    If l < 2 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (Pox_CharIsHex(c) or (c = '-') or (c = '#')) then exit
        If (i > 1) and (c = '-') then exit
        If (i > 2) and (c = '#') then exit
        If (c = '#') then Inc(sCount)
    Next
    
    result = TRUE
    
    If sCount <> 1 then result = FALSE        
End

Function Pox_StringIsBinary(s: String) : Boolean ; export
Var
    l, i: Integer
    c: String
    sCount: Integer = 0 // symbolCount
Begin
    result = FALSE
    l = Length(s)
    If l < 2 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (Pox_CharIsNumeric(c) or (c = '%')) then exit
        If (i > 1) and (c = '%') then exit
        If (c = '%') then Inc(sCount)
    Next
    
    result = TRUE
    
    If sCount <> 1 then result = FALSE        
End

Procedure Pox_Error(s: String)
Begin
    Inc(poxErrors)
    If SUPPRESS_ERRORS then exit
    MessageBox(s, "Poxml")
End

Procedure Pox_ExpectedError(s: String)
Begin
    Pox_CodeError("Expected "+s)
End

Procedure Pox_UnexpectedError(s: String)
Begin
    Pox_CodeError("Unexpected "+s)
End

Procedure Pox_CodeError(s: String)
Begin

    s = "Error: "+s+" at line "+((ItemInList(poxSourceList, poxCurrLine)+1))
    s = s + ":"+ToString(poxLexer_index0)
    s = s + Chr(10) + Replace(poxCurrLine.s, Chr(9), ' ', FALSE) // tabs -> spaces
    
    // with a fixed-width font, this will display a caret at the error position
    //s = s + Chr(10) + RepStr(' ', iniLexer_index0)+'^'
    
    Pox_Error(s)

End

Function HexStringToInt(s: String): Integer
Var
    l, i, c: Integer
    polarity: Integer
    start: Integer
    total: Integer = 0
Begin
    l = Length(s)
    s = Upper(s)
    
    // Remove -, # (note assumes valid hex string, e.g. -#FF)
    If Left(s, 1) = '-' then
        polarity = -1
        start = 3
    Else
        polarity = 1
        start = 2
    Endif
    
    For i = start to l
        c = Asc(Mid(s, i, 1))
        If (c >= 48) and (c <= 57) then c = c - 48 // 0-9
        If (c >= 65) and (c <= 70) then c = c + 10 - 65 // A-F
        
        If (l - i) > 0 then
            total = total + (c * (16^(l - i)))
        Else
            total = total + c
        Endif         
    
    Next
    
    result = total * polarity   

End

Function BinStringToInt(s: String): Integer
Var
    l, i, c: Integer
    total: Integer = 0
Begin
    l = Length(s)

    For i = 2 to l
        c = Asc(Mid(s, i, 1))
        If (c = 48) then c = 0 else c = 1
        
        If (l - i) > 0 then
            total = total + (c * (2^(l - i)))
        Else
            total = total + c
        Endif         
    
    Next
    
    result = total

End

Function Pox_SpecialChrs(s: String): String
Begin
    //&amp; & &lt; < &gt; > &quot; " &apos; '
    s = Replace(s,"&quot;", '"', TRUE)
    s = Replace(s,"&apos;", "'", TRUE)
    s = Replace(s,"&lt;",   "<", TRUE)
    s = Replace(s,"&gt;",   ">", TRUE)
    s = Replace(s,"&amp;", "&",  TRUE)
    result = s
End

Begin
End